from typing import Union

from .constants import (
    NOTE_TO_MIDI,
    ROOT_TO_MIDI,
    NOTE_EQUIVALENTS,
    MIDI_TO_FREQ,
    MIDI_TO_NOTE,
    NOTE_TO_FREQ,
)


def sharp_to_flat(note_name: str) -> str:
    """
    Translate a sharp note to a flat note, since we only use flats internally
    Args:
        note_name: A valid note name from `midi_utils.constants.NOTE_TO_MIDI`
              or a midi note between 0 and 127.
    """
    if "#" in note_name:
        try:
            sharp_note = NOTE_EQUIVALENTS[note_name[:2]]
        except KeyError:
            raise ValueError(f"Invalid note: {note_name}")
        return sharp_note + note_name[2:]
    return note_name


def note_to_midi(note_name: Union[str, int]) -> int:
    """
    Convert a note name (eg: "C3") into its corresponding midi note value (eg: 60).
    If passed a valid midi note, it will simply return it.
    Args:
        note_name: A valid note name from `midi_utils.constants.NOTE_TO_MIDI`
    """
    if isinstance(note_name, str):
        try:
            midi_note = int(note_name)
        except ValueError:
            note_name = str(note_name).upper()
            note_name = sharp_to_flat(note_name)
            if note_name in NOTE_TO_MIDI:
                midi_note = NOTE_TO_MIDI[note_name]
            else:
                raise ValueError(f'"{note_name}" is not a valid note.')
    else:
        midi_note = int(note_name)
    if midi_note < 0 or midi_note > 127:
        raise ValueError(
            f'"{midi_note}" is not a valid midi note. Must be between 0 and 127'
        )
    return midi_note


def note_to_freq(note_name: str) -> int:
    """
    Convert a note name (eg: "A4") into its corresponding frequency (eg: 440).
    Args:
        note_name: A valid note name from `midi_utils.constants.NOTE_TO_MIDI`
    """
    try:
        return NOTE_TO_FREQ[note_name]
    except KeyError:
        raise ValueError(f"Invalid note name: {note_name}")


def midi_to_note(midi_note_num: int) -> str:
    """
    Convert a midi note number (eg: 60) into its corresponding note name (eg: "C3").
    If passed a valid midi note, this function will simply return it.

    Args:
        midi_note_num: A midi note number between 0 and 127
    """
    try:
        return MIDI_TO_NOTE[midi_note_num]
    except KeyError:
        raise ValueError(f"Invalid midi note num: {midi_note_num}")


def midi_to_freq(midi_note_num: int) -> float:
    """
    Convert a midi note number (eg: 81) into its corresponding note frequency (eg: 440).
    Args:
        midi_note_num: A midi note number between 0 and 127
    """
    try:
        return MIDI_TO_FREQ[midi_note_num]
    except KeyError:
        raise ValueError(f"Invalid midi note num: {midi_note_num}")


def midi_to_octave(midi_note_num: int, octave: int) -> int:
    """
    Raise a midi note up or down an one or more octaves

    Args:
        midi_note_num: A valid midi note between 0 and 127.
        octave: the number of octaves to adjust the note up (1) or down (-1)
    """
    midi_note_num += octave * 12
    if midi_note_num < 0 or midi_note_num > 127:
        raise ValueError(f"midi note num {midi_note_num} out of valid range.")
    return midi_note_num


def root_to_midi(root: str) -> int:
    """
    Convert a root name (eg: "C") into its corresponding midi note  .
    If passed a valid midi note value, this function will simply return it.

    Args:
        root: A valid root name from `midi_utils.constants.NOTES`
              or a midi note between 0 and 11.
    """
    if isinstance(root, str):
        root = root.upper()
        if root in ROOT_TO_MIDI:
            return ROOT_TO_MIDI[root]
        raise ValueError(
            f'"{root}" is not a valid root name. Choose from: {", ".join(ROOT_TO_MIDI.keys())}'
        )
    if root not in ROOT_TO_MIDI.values():
        raise ValueError(
            f'"{root}" is not a valid midi root. Choose from: {", ".join(ROOT_TO_MIDI.values())}'
        )
    return root


def freq_to_octave(freq: float, octave: int) -> float:
    """
    Adjust a frequency up or down by octaves
    """
    if octave < 0:
        return freq * (1.0 / (abs(octave) * 2.0))
    return freq * (octave * 2.0)


def bpm_to_time(
    bpm: float = 120.00, count: Union[str, int, float] = 1, time_sig: str = "4/4"
):
    """
    Take a bpm, note count, and timesig and return a length in seconds
    """
    if isinstance(count, str):
        if "/" in count:
            numerator, denominator = count.split("/")
            count = float(numerator) / float(denominator)
    time_segs = time_sig.split("/")
    return (60.00 / float(bpm)) * float(time_segs[0]) * float(count) * 1000.0


def rescale(x, range_x, range_y, sig_digits=3):

    # Figure out how 'wide' each range is
    x_span = max(range_x) - min(range_x)
    y_span = max(range_y) - min(range_y)

    # Compute the scale factor between left and right values
    scale_factor = float(y_span) / float(x_span)

    return round(min(range_y) + (x - min(range_x)) * scale_factor, sig_digits)



from typing import Union, List

from .constants import SCALES
from .utils import note_to_midi, root_to_midi


def midi_scale(
    key: Union[str, int],
    scale: Union[List[int], str],
    min_note: int = 0,
    max_note: int = 127,
    **kwargs,
) -> List[int]:
    """
    Generate a scale of midi notes.

    Args:
        key: A valid root name from `midi_utils.constants.ROOTS`
             or a midi note between 0 and 11.
        scale: A valid scale name from `midi_utils.constants.SCALES`
               or a list of integers, eg [0, 4, 7]
        min_note: The lowest note in the scale;
                  A valid note name from `midi_utils.constants.NOTES`
                  or a midi note between 0 and 127
        max_note: The lowest note in the scale;
                  A valid note name from `midi_utils.constants.NOTES`
                  or a midi note between 0 and 127

    """
    if min_note > max_note:
        raise ValueError(
            'The min_note "%s" is larger than the max_note "%s"' % min_note, max_note
        )
    # convert args
    key = root_to_midi(key)
    min_note = note_to_midi(min_note)
    max_note = note_to_midi(max_note)

    # lookup scale by name
    if isinstance(scale, str):
        scale = scale.upper().replace(" ", "_").strip()
        if scale in SCALES:
            scale = SCALES[scale]
        else:
            raise ValueError(
                f'"{scale}" is not a valid scale name. Choose from: {", ".join(SCALES.keys())}'
            )

    # position it according to the key
    key_scale = [s + key for s in scale]

    # stretch from min note to max note
    return [
        x + (12 * j)
        for j in range(12)
        for x in key_scale
        if x + (12 * j) >= min_note and x + (12 * j) <= max_note
    ]


def map_to_midi_scale(x, scale: List[int], min_x: int, max_x: int, **kwargs) -> int:
    """
    Map an arbitrary value (`x`) to a provided midi `scale` given the `min_x` and `max_x` values.
    This is useful for generated data sonifications in a particular key.

    Args:
        x: An integer to map to a midi note
        scale: A scale generated by `midi_utils.midi_scale`
        min_x: The lowest possible value of `x`
        max_x: The highest possible value of `x`
    """
    scale_length = len(scale) - 1
    scale_idx = int(round(((x - min_x) * scale_length) / (max_x - min_x)))
    return scale[scale_idx]



# ADSR generator
from platform import release
from typing import Generator

from .utils import rescale


class ADSR:
    def __init__(
        self,
        attack: float = 0.2,
        decay: float = 0.2,
        sustain: float = 0.7,
        release: float = 0.3,
        samples: int = 20,
        attack_min_level: float = 0.0,
        attack_max_level: float = 1.0,
        sig_digits: int = 2,
        is_zero_indexed: bool = True,
    ):
        self.attack = (
            attack  # the percentage of samples it takes to reach the peak amplitude
        )
        self.decay = decay  # the percentage of samples it takes to descend to the sustain amplitude
        self.sustain_level = sustain  # the amplitude at the sustain level
        self.release = release  # the percentage of samples it takes to reach 0, from the end of the note
        self.sustain = 1.0 - (self.attack + self.decay + self.release)
        # the percentage of samples it stays at the sustain level.
        # this is calculated by subtracting the sum of attack, decay, and release from 1
        if self.sustain < 0:
            raise ValueError(
                "The sum of attack, decay, and release percentages are greater than 1"
            )
        self.samples = samples  # the number of samples in the note
        self.attack_min_level = attack_min_level
        self.attack_max_level = attack_max_level
        self.sig_digits = sig_digits  # what to round values to
        self.is_zero_indexed = is_zero_indexed
        self.idx = -1

    def get_value(self, sample: int):
        """
        Get value of envelope at sample i
        """
        # since we'll be using this with a list index, we'll have to handle cases
        # where the index is 0
        if not self.is_zero_indexed:
            sample += 1
        sample_per = float(sample) / float(self.samples)
        val = 0  # default

        # attack phase
        if sample_per <= self.attack:
            if self.attack == 0:
                val = 1
            else:
                val = rescale(
                    sample_per / self.attack,
                    range_x=[0, 1],
                    range_y=[self.attack_min_level, self.attack_max_level],
                )

        # decay phase
        elif (sample_per - self.attack) < self.decay:
            # descend from 1 to sustain level
            if self.decay > 0:
                decay_ratio = (sample_per - self.attack) / self.decay
            else:
                decay_ratio = 0
            val = 1 - (decay_ratio * (1 - self.sustain_level))

        # sustain phase
        elif (sample_per - self.attack - self.decay) < self.sustain:
            # remain at sustain level
            val = self.sustain_level

        # release phase
        else:
            # descend from sustain level to 0
            if self.release > 0:
                release_ratio = (
                    sample_per - self.attack - self.decay - self.sustain
                ) / self.release
            else:
                release_ratio = 0
            val = (1 - release_ratio) * self.sustain_level
        # never go below zero
        return max(0.0, round(val, self.sig_digits))

    def __iter__(self) -> Generator[float, None, None]:
        for i in range(0, self.samples):
            yield self.get_value(i)

    def __next__(self):
        self.idx += 1
        if self.idx > self.samples:
            raise StopIteration
        return self.get_value(self.idx)



from functools import partial
import random
import copy
from typing import List, Union, Optional
import logging

from .utils import bpm_to_time, note_to_midi
from .chord import midi_chord

log = logging.getLogger(__name__)


def _random_chance(prob):
    return random.randint(1, 100) <= prob


def _oddeven(notes, even_first=False, **kwargs):
    odd = []
    even = []
    for i, n in enumerate(notes, start=1):
        if i % 2:
            even.append(n)
        else:
            odd.append(n)
    if even_first:
        return even + odd
    return odd + even


def oddeven(notes: List[int], **kwargs) -> List[int]:
    """
    odd indexed notes followed by even indexed notes
    """
    return _oddeven(notes)


def evenodd(notes: List[int], **kwargs) -> List[int]:
    """
    even indexed notes followed by odd indexed notes
    """
    return _oddeven(notes, even_first=True)


def _shift_n(notes: List[int], n: int = 1, **kwargs) -> List[int]:
    """
    switch the last n notes with the first n notes
    """
    n = min(n, len(notes))
    return notes[n:] + notes[:n]


def thumb_up(notes: List[int], **kwargs) -> List[int]:
    """
    alternate between the lowest note and an ascending scale
    """
    sorted_notes = list(sorted(notes))
    ordered_notes = []
    for note in sorted_notes[1:]:
        ordered_notes.extend([sorted_notes[0], note])
    return ordered_notes


def thumb_down(notes: List[int], **kwargs) -> List[int]:
    """
    alternate between the lowest note and a descending scale
    """
    sorted_notes = list(reversed(sorted(notes)))
    ordered_notes = []
    for note in sorted_notes[:-1]:
        ordered_notes.extend([sorted_notes[-1], note])
    return ordered_notes


def thumb_updown(notes: List[int], **kwargs) -> List[int]:
    """
    thumb up then down, without repeating highest and lowest notes.
    """
    return thumb_up(notes, **kwargs) + thumb_down(notes, **kwargs)[2:-2]


def thumb_downup(notes: List[int], **kwargs) -> List[int]:
    """
    thumb down then up, without repeating lowest and highest notes
    """
    return thumb_down(notes, **kwargs) + thumb_up(notes, **kwargs)[2:-2]


def _get_middle_note(notes, **kwargs):
    """
    find the middle element of a list of notes
    """
    middle = float(len(notes)) / 2
    if middle % 2 != 0:
        return notes[int(middle - 0.5)]
    return (notes[int(middle)], notes[int(middle - 1)])


def _gen_middle(notes, middle_note, overlapping=False, **kwargs):
    """
    generate a "middle" pattern
    """
    ordered_notes = []
    for note in notes:
        if note == middle_note:
            if overlapping:
                ordered_notes.extend([middle_note, note])
        else:
            ordered_notes.extend([middle_note, note])
    return ordered_notes


def middle_up(notes: List[int], overlapping=False, **kwargs) -> List[int]:
    """
    alternate between the middle note and an ascending scale
    """
    middle_note = _get_middle_note(notes)
    return _gen_middle(sorted(notes), middle_note, overlapping, **kwargs)


def middle_down(notes: List[int], overlapping=False, **kwargs) -> List[int]:
    """
    alternate between the middle note and a descending scale
    """
    middle_note = _get_middle_note(notes)
    return _gen_middle(sorted(notes, reverse=True), middle_note, overlapping, **kwargs)


def middle_updown(notes: List[int], **kwargs) -> List[int]:
    """
    middle up then down, without repeating notes.
    """
    return middle_up(notes, **kwargs) + middle_down(notes, **kwargs)[2:-2]


def middle_downup(notes: List[int], **kwargs) -> List[int]:
    """
    middle down then up, without repeating notes
    """
    return middle_down(notes, **kwargs) + middle_up(notes, **kwargs)[2:-2]


def pinky_up(notes: List[int], **kwargs) -> List[int]:
    """
    alternate between the highest note and an ascending scale
    """
    sorted_notes = list(sorted(notes))
    ordered_notes = []
    for note in sorted_notes[:-1]:
        ordered_notes.extend([sorted_notes[-1], note])
    return ordered_notes


def pinky_down(notes: List[int], **kwargs) -> List[int]:
    """
    alternate between the highest note and a descending scale
    """
    sorted_notes = list(reversed(sorted(notes)))
    ordered_notes = []
    for note in sorted_notes[1:]:
        ordered_notes.extend([sorted_notes[-1], note])
    return ordered_notes


def pinky_updown(notes: List[int], **kwargs) -> List[int]:
    """
    pinky up then down, without repeating highest and lowest notes.
    """
    return pinky_up(notes, **kwargs) + pinky_down(notes, **kwargs)[2:-2]


def pinky_downup(notes: List[int], **kwargs) -> List[int]:
    """
    pinky down then up, without repeating lowest and highest notes
    """
    return pinky_down(notes, **kwargs) + pinky_up(notes, **kwargs)[2:-2]


def random_shuffle(notes: List[int], **kwargs) -> List[int]:
    """
    randomly shuffle notes
    """
    random.shuffle(notes)
    return notes


def random_octaves(
    notes: List[int], octaves: List[int] = [1], prob_octave: int = 33, **kwargs
) -> List[int]:
    """
    randomly apply octaves to a sequence
    """
    new_notes = []
    for note in notes:
        if _random_chance(prob_octave):
            octave = random.choice(octaves)
            note += octave * 12
        new_notes.append(note)
    return new_notes


def random_silence(notes, prob_silence: int = 33, **kwargs) -> List[int]:
    """
    randomly remove notes from a sequence
    """
    new_notes = []
    for note in notes:
        if _random_chance(prob_silence):
            note = 0
        new_notes.append(note)
    return new_notes


def three_two(notes, **kwargs) -> List[int]:
    """
    3 notes then 2 of silence
    """
    breaks = range(0, len(notes), 5)
    for b in breaks[1:]:
        notes[b - 2 : b] = [0, 0]
    return notes


def two_three(notes, **kwargs) -> List[int]:
    """
    2 notes then 3 of silence
    """
    breaks = range(0, len(notes), 5)
    for b in breaks[1:]:
        notes[b - 3 : b] = [0, 0, 0]
    return notes


def even_off(notes, prob_even_off: int = 33, **kwargs) -> List[int]:
    """
    turn even notes randomly off
    """
    new_notes = []
    for i, note in enumerate(notes, 1):
        if i % 2 == 0 and _random_chance(prob_even_off):
            note = 0
        new_notes.append(note)
    return new_notes


def odd_off(notes, prob_odd_off: int = 29, **kwargs) -> List[int]:
    """
    turn odd notes randomly off
    """
    new_notes = []
    for i, note in enumerate(notes, 1):
        if i % 2 != 0 and _random_chance(prob_odd_off):
            note = 0
        new_notes.append(note)
    return new_notes


def down(x, **kwargs):
    return sorted(x, reverse=True)


def up(x, **kwargs):
    return sorted(x, reverse=False)


def downup(x, **kwargs):
    return list(reversed(sorted(x))) + list(sorted(x))[1:-1]


def updown(x, **kwargs):
    return list(sorted(x)) + list(reversed(sorted(x)))[1:-1]


def repeat_start(x, **kwargs):
    repeat_start_range = kwargs.get("repeat_start__range", [0])
    n = kwargs.get("repeat_start__n", repeat_start_range)
    return (x[0] * n) + x[1:]


def repeat_end(x, **kwargs):
    repeat_end_range = kwargs.get("repeat_end_range", [0])
    n = kwargs.get("repeat_end_n", repeat_end_range)
    return (x[0] * n) + x[1:]


STYLES = {
    "down": down,
    "up": up,
    "downup": downup,
    "updown": updown,
    "oddeven": oddeven,
    "evenodd": evenodd,
    "repeat_start": repeat_start,
    "repeat_end": repeat_end,
    "shift_1": partial(_shift_n, n=1),
    "shift_2": partial(_shift_n, n=2),
    "shift_3": partial(_shift_n, n=3),
    "shift_4": partial(_shift_n, n=4),
    "shift_5": partial(_shift_n, n=5),
    "shift_6": partial(_shift_n, n=6),
    "shift_7": partial(_shift_n, n=7),
    "shift_8": partial(_shift_n, n=8),
    "middle_up": middle_up,
    "middle_down": middle_down,
    "middle_updown": middle_updown,
    "middle_downup": middle_downup,
    "thumb_up": thumb_up,
    "thumb_down": thumb_down,
    "thumb_updown": thumb_updown,
    "thumb_downup": thumb_downup,
    "pinky_up": pinky_up,
    "pinky_down": pinky_down,
    "pinky_updown": pinky_updown,
    "pinky_downup": pinky_downup,
    "random": random_shuffle,
    "random_shuffle": random_shuffle,
    "random_octaves": random_octaves,
    "random_silence": random_silence,
    "three_two": three_two,
    "two_three": two_three,
    "odd_off": odd_off,
    "even_off": even_off,
}


def apply_style(notes: List[int], style: str = "down", **kwargs) -> List[int]:
    """
    Apply a style to a list of notes
    """
    if style not in STYLES:
        raise ValueError(f'Style "{style}" is not valid. Choose from:\n\n{", ".join(STYLES.keys())}')
    return STYLES.get(style)(notes, **kwargs)


def apply_styles(x: List[int], styles: List[str], **kwargs):
    """
    Apply a list of styles to a list of notes in order from left to right
    """
    # apply multiple styles, in order
    for style in styles:
        notes = apply_style(x, style, **kwargs)
    return notes


def _check_duration(total_time, duration):
    if duration:
        if duration >= total_time:
            return True
    return False


def _gen_silence(duration):
    return {
        "note": 0,
        "type": "silence",
        "duration": duration,
        "velocity": 0,
    }


def midi_arp(
    notes: List[int] = [],  # arbitrary list of notes to arpeggiate
    root: str = "A3",  # root note of chord
    chord: str = "min6_9",  # chord name,
    inversions: List[int] = [],  # inversions list
    stack: int = 0,  # stack a chord up or down
    styles: List[str] = ["down"],
    octaves: List[int] = [0],  # a list of octaves to add to the notes (eg: [-1, 2])
    velocities: List[int] = [100],
    # a list of velocities to apply across notes,
    # velocities are retrieved using a modulo so
    # this can be any duration and will be applied
    # in order
    beat_bpm: float = 131.0,  # bpm to use when determining beat length
    beat_count: Union[float, int, str] = "1/16",  # count of one beat of the arp
    beat_time_sig: str = "4/4",  # time signature of arp
    beat_duration: Optional[float] = None,  # the time of the note in ms
    note_bpm: float = 131.0,  # bpm to use when determining note duration
    note_count: Union[
        float, int, str
    ] = "3/64",  # count of note legth (should be less than or equat to beat count)
    note_time_sig: str = "4/4",  # time signature of arp
    note_duration: Optional[float] = None,  # arbitrary duration of note in ms
    start_bpm: float = 131.0,  # bpm to use when determining the start of the arp and adds silence at the beginning
    start_count: Union[float, int, str] = 0,  # the start beat count
    start_time_sig: str = "4/4",  # time signature to use when determining start
    start_duration: Optional[
        float
    ] = None,  # the amount of silence to add at the beginning in ms
    duration_bpm: float = 131.0,  # bpm to use when determining note duration
    duration_count: Union[float, int, str] = "16",  # the duration beat count
    duration_time_sig: str = "4/4",  # time signature to use when determining duration
    duration: Optional[float] = None,  # the total duration of the pattern in ms
    loops: Optional[int] = None,
    **kwargs,
):
    """
    Given a list of chord notes, a style name (eg: up, down), and a note count, generate an arpeggiated sequence of notes.
    """

    # generate notes
    if len(notes):
        chord_notes = [note_to_midi(n) for n in notes]
    else:
        chord_notes = midi_chord(root, chord, inversions, stack, **kwargs)

    # create list of notes including octaves
    chord_notes = list(
        set(chord_notes + [o * 12 + n for o in octaves for n in chord_notes])
    )
    # determine beat_duration
    if not beat_duration:
        beat_duration = bpm_to_time(beat_bpm, beat_count, beat_time_sig)

    # determine note_duration
    if not note_duration:
        note_duration = min(
            bpm_to_time(note_bpm, note_count, note_time_sig), beat_duration
        )
    else:
        note_duration = copy.copy(beat_duration)

    # clamp note_duration to beat_duration
    if note_duration > beat_duration:
        log.warning(
            f"WARNING: note_duration {note_duration} is greater than {beat_duration}, "
        )
        note_duration = copy.copy(beat_duration)

    # determine total duration
    if not duration:
        duration = bpm_to_time(duration_bpm, duration_count, duration_time_sig)

    # add silence
    if not start_duration:
        start_duration = bpm_to_time(start_bpm, start_count, start_time_sig)

    if start_duration and start_duration > 0:
        yield _gen_silence(start_duration)

    # keep track of time and loops
    total_time = 0.0
    n_loops = 0

    #
    while True:
        # apply the style /  order
        for i, note in enumerate(apply_styles(chord_notes, styles, **kwargs)):

            # determine the velocity
            velocity = velocities[i % len(velocities)]
            # TODO: randomize velocity

            # TODO: randomize note length

            yield {
                "note": note,
                "type": "note" if note != 0 else "silence",
                "duration": note_duration,
                "velocity": velocity,
            }
            silence_duration = beat_duration - note_duration
            if silence_duration > 0:
                yield _gen_silence(silence_duration)

            # check for max duration after each beat if loops are not set.
            total_time += beat_duration
            if not loops and _check_duration(total_time, duration):
                break

        # check for max duration after breaking out of the notes loop if loops are not set
        if not loops and _check_duration(total_time, duration):
            break

        # check for loops if set
        n_loops += 1
        if loops and n_loops >= loops:
            break



from typing import List, Union

from midi_utils.constants import CHORDS
from midi_utils.utils import note_to_midi


def lookup_chord(root: Union[str, int], chord: str, **kwargs) -> List[int]:
    """
    Given a root and a chord name return a list of chord midi notes
    """
    midi_root = note_to_midi(root)
    base_notes = CHORDS.get(chord.upper().strip())
    if not base_notes:
        raise ValueError(
            f'Invalid chord name: {chord}. Choose from:{",".join(CHORDS.keys())}'
        )
    return [midi_root + note for note in base_notes]


def invert_chord(
    chord_notes: List[int], inversions: List[int] = [], **kwargs
) -> List[str]:
    """
    Given a list of chord notes and a list of inversions, return an inverted chord
    """
    if not len(inversions):
        return chord_notes

    inverted_chord = []
    for i, note in enumerate(chord_notes):
        inv = inversions[i % len(inversions)]
        note += inv * 12
        inverted_chord.append(note)
    return inverted_chord


def stack_chord(chord_notes: List[int], stack: int = 0, **kwargs):
    """
    Given a list of chord notes and a negative/positive integer, stack a chord and select its unique notes.
    """
    stacked_chord = [(stack * 12) + note for note in chord_notes]
    return list(sorted(list(set(stacked_chord + chord_notes))))


def midi_chord(
    root: Union[str, int], chord="MIN_6_9", inversions=[], stack=0, **kwargs
):
    """
    Given a chord root, chord name, list of inversions, and a number and direction to stack, generate a chord.
    """
    chord_notes = lookup_chord(root, chord)
    return stack_chord(invert_chord(chord_notes, inversions), stack)



from setuptools import setup, find_packages


config = {
    "name": "midi-utils",
    "version": "1.0.0",
    "packages": find_packages(),
    "test_requires": [
        "pytest",
    ],
    "author": "Brian Abelson",
    "author_email": "hey@gltd.email",
    "description": "midi-utils provides a couple of simple tools to generate midi notes in different scales",
    "url": "https://gitlab.com/gltd/midi-utils",
}

setup(**config)



from .scale import midi_scale, map_to_midi_scale
from .chord import midi_chord
from .arp import midi_arp
from .utils import (
    note_to_freq,
    note_to_midi,
    root_to_midi,
    note_to_freq,
    midi_to_freq,
    midi_to_note,
    midi_to_octave,
    freq_to_octave,
    sharp_to_flat,
)
from .adsr import ADSR
from .constants import *
